///////////////////////////////////////////////////////////////////////////////
//
/// \file       block_header_decoder.c
/// \brief      Decodes Block Header from .lzma files
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "common.h"
#include "check.h"


struct lzma_coder_s {
	lzma_options_block *options;

	enum {
		SEQ_FLAGS_1,
		SEQ_FLAGS_2,
		SEQ_COMPRESSED_SIZE,
		SEQ_UNCOMPRESSED_SIZE,
		SEQ_FILTER_FLAGS_INIT,
		SEQ_FILTER_FLAGS_DECODE,
		SEQ_CRC32,
		SEQ_PADDING
	} sequence;

	/// \brief      Position in variable-length integers
	size_t pos;

	/// \brief      CRC32 of the Block Header
	uint32_t crc32;

	lzma_next_coder filter_flags_decoder;
};


static lzma_ret
block_header_decode(lzma_coder *coder, lzma_allocator *allocator,
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out lzma_attribute((unused)),
		size_t *restrict out_pos lzma_attribute((unused)),
		size_t out_size lzma_attribute((unused)),
		lzma_action action lzma_attribute((unused)))
{
	while (*in_pos < in_size)
	switch (coder->sequence) {
	case SEQ_FLAGS_1:
		// Check that the reserved bit is unset.
		if (in[*in_pos] & 0x40)
			return LZMA_DATA_ERROR;

		// Number of filters: we prepare appropriate amount of
		// variables for variable-length integer parsing.
		for (int i = (int)(in[*in_pos] & 0x07) - 1; i >= 0; --i)
			coder->options->filters[i].id = 0;

		// End of Payload Marker flag
		coder->options->has_eopm = (in[*in_pos] & 0x08) != 0;

		// Compressed Size: Prepare for variable-length integer
		// parsing if it is known.
		if (in[*in_pos] & 0x10)
			coder->options->compressed_size = 0;

		// Uncompressed Size: the same.
		if (in[*in_pos] & 0x20)
			coder->options->uncompressed_size = 0;

		// Is Metadata Block flag
		coder->options->is_metadata = (in[*in_pos] & 0x80) != 0;

		// We need at least one: Uncompressed Size or EOPM.
		if (coder->options->uncompressed_size == UINT64_MAX
				&& !coder->options->has_eopm)
			return LZMA_DATA_ERROR;

		// Update header CRC32.
		coder->crc32 = lzma_crc32(in + *in_pos, 1, coder->crc32);

		++*in_pos;
		coder->sequence = SEQ_FLAGS_2;
		break;

	case SEQ_FLAGS_2:
		// Check that the reserved bits are unset.
		if (in[*in_pos] & 0xE0)
			return LZMA_DATA_ERROR;

		// Get the size of Header Padding.
		coder->options->alignment = in[*in_pos] & 0x1F;

		coder->crc32 = lzma_crc32(in + *in_pos, 1, coder->crc32);

		++*in_pos;
		coder->sequence = SEQ_COMPRESSED_SIZE;
		break;

	case SEQ_COMPRESSED_SIZE:
		if (coder->options->compressed_size != UINT64_MAX) {
			// Store the old input position to be used when
			// updating coder->header_crc32.
			const size_t in_start = *in_pos;

			const lzma_ret ret = lzma_vli_decode(
					&coder->options->compressed_size,
					&coder->pos, in, in_pos, in_size);

			coder->crc32 = lzma_crc32(in + in_start,
					*in_pos - in_start, coder->crc32);

			if (ret != LZMA_STREAM_END)
				return ret;

			coder->pos = 0;
			coder->sequence = SEQ_UNCOMPRESSED_SIZE;
			break;
		}

		coder->sequence = SEQ_UNCOMPRESSED_SIZE;

	// Fall through

	case SEQ_UNCOMPRESSED_SIZE:
		if (coder->options->uncompressed_size != UINT64_MAX) {
			const size_t in_start = *in_pos;

			const lzma_ret ret = lzma_vli_decode(
					&coder->options->uncompressed_size,
					&coder->pos, in, in_pos, in_size);

			coder->crc32 = lzma_crc32(in + in_start,
					*in_pos - in_start, coder->crc32);

			if (ret != LZMA_STREAM_END)
				return ret;

			coder->pos = 0;
			coder->sequence = SEQ_FILTER_FLAGS_INIT;
			break;
		}

	// Fall through

	case SEQ_FILTER_FLAGS_INIT: {
		// Check if we have parsed the whole List of Filter Flags.
		if (coder->options->filters[coder->pos].id == UINT64_MAX) {
			coder->pos = 0;
			coder->sequence = SEQ_CRC32;
			break;
		}

		const lzma_ret ret = lzma_filter_flags_decoder_init(
				&coder->filter_flags_decoder, allocator,
				&coder->options->filters[coder->pos]);
		if (ret != LZMA_OK)
			return ret;

		coder->sequence = SEQ_FILTER_FLAGS_DECODE;
	}

	// Fall through

	case SEQ_FILTER_FLAGS_DECODE: {
			const size_t in_start = *in_pos;

		const lzma_ret ret = coder->filter_flags_decoder.code(
				coder->filter_flags_decoder.coder,
				allocator, in, in_pos, in_size,
				NULL, NULL, 0, LZMA_RUN);

		coder->crc32 = lzma_crc32(in + in_start,
				*in_pos - in_start, coder->crc32);

		if (ret != LZMA_STREAM_END)
			return ret;

		++coder->pos;
		coder->sequence = SEQ_FILTER_FLAGS_INIT;
		break;
	}

	case SEQ_CRC32:
		if (coder->options->has_crc32) {
			if (in[*in_pos] != ((coder->crc32 >> (coder->pos * 8))
					& 0xFF))
				return LZMA_DATA_ERROR;

			if (++coder->pos == 4) {
				coder->pos = 0;
				coder->sequence = SEQ_PADDING;
			}

			++*in_pos;
			break;
		}

		coder->sequence = SEQ_PADDING;

	// Fall through

	case SEQ_PADDING:
		if (coder->pos < coder->options->alignment) {
			if (in[*in_pos] != 0x00)
				return LZMA_DATA_ERROR;

			++coder->pos;
			++*in_pos;
			break;
		}

		return LZMA_STREAM_END;

	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}


static void
block_header_decoder_end(lzma_coder *coder, lzma_allocator *allocator)
{
	lzma_next_coder_end(&coder->filter_flags_decoder, allocator);
	lzma_free(coder, allocator);
	return;
}


extern lzma_ret
lzma_block_header_decoder_init(lzma_next_coder *next,
		lzma_allocator *allocator, lzma_options_block *options)
{
	if (next->init != &lzma_block_header_decoder_init) {
		lzma_next_coder_end(next, allocator);

		next->coder = lzma_alloc(sizeof(lzma_coder), allocator);
		if (next->coder == NULL)
			return LZMA_MEM_ERROR;

		next->coder->filter_flags_decoder = LZMA_NEXT_CODER_INIT;
	}

	// We don't care about options->check and options->has_crc32.

	// Assume that Compressed Size and Uncompressed Size are unknown.
	options->compressed_size = UINT64_MAX;
	options->uncompressed_size = UINT64_MAX;

	// See SEQ_FLAGS_1 in block_header_decode() for reasoning of this.
	for (size_t i = 0; i < 8; ++i) {
		options->filters[i].id = UINT64_MAX;
		options->filters[i].options = NULL;
	}

	next->coder->options = options;
	next->coder->sequence = SEQ_FLAGS_1;
	next->coder->pos = 0;
	next->coder->crc32 = 0;

	// next->coder->filter_flags_decoder is initialized when allocated,
	// and before use.

	next->init = &lzma_block_header_decoder_init;
	next->code = &block_header_decode;
	next->end = &block_header_decoder_end;

	return LZMA_OK;
}


extern LZMA_API lzma_ret
lzma_block_header_decoder(lzma_stream *strm, lzma_options_block *options)
{
	lzma_ret ret;

	ret = lzma_strm_init(strm);
	if (ret != LZMA_OK)
		return ret;

	strm->internal->supported_actions[LZMA_RUN] = true;

	ret = lzma_block_header_decoder_init(
			&strm->internal->next, strm->allocator, options);
	if (ret != LZMA_OK)
		lzma_end(strm);

	return ret;
}
