///////////////////////////////////////////////////////////////////////////////
//
/// \file       filter_flags_decoder.c
/// \brief      Decodes a Filter Flags field
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "filter_flags_common.h"
#include "lzma_encoder.h"


/// \brief      Calculates the size of the Filter Properties field
static int32_t
get_properties_size(const lzma_options_filter *options)
{
	int32_t ret;

	switch (options->id) {
	case LZMA_FILTER_SPARSE:
		// Reject too big repeat buffer size.
		if (((lzma_options_sparse *)(options->options))->size
				> LZMA_FILTER_PROPERTIES_SIZE_MAX)
			ret = -1;
		else
			ret = ((lzma_options_sparse *)(options->options))
					->size;
		break;

	case LZMA_FILTER_LZMA:
		ret = 2;
		break;

	case LZMA_FILTER_COPY:
	case LZMA_FILTER_SUBBLOCK:
	case LZMA_FILTER_X86:
	case LZMA_FILTER_POWERPC:
	case LZMA_FILTER_IA64:
	case LZMA_FILTER_ARM:
	case LZMA_FILTER_ARMTHUMB:
	case LZMA_FILTER_SPARC:
		// No Filter Properties
		ret = 0;
		break;

	default:
		// Unknown filter
		ret = -1;
	}

	return ret;
}


extern LZMA_API size_t
lzma_filter_flags_size(const lzma_options_filter *options)
{
	// Get size of Filter Properties.
	const int32_t prop_size = get_properties_size(options);
	if (prop_size == -1)
		return 0;

	// Size of Filter ID field if it exists.
	size_t id_size;
	size_t prop_size_size;
	if (options->id < 0xE0
			&& (uint64_t)(prop_size) == options->id / 0x20) {
		// ID and Size of Filter Properties fit into Misc.
		id_size = 0;
		prop_size_size = 0;

	} else {
		// At least Filter ID is stored using the External ID field.
		id_size = lzma_vli_size(options->id);
		if (id_size == 0)
			return 0;

		if (prop_size <= 30) {
			// Size of Filter Properties fits into Misc still.
			prop_size_size = 0;
		} else {
			// The Size of Filter Properties field is used too.
			prop_size_size = lzma_vli_size((uint64_t)(prop_size));
			if (prop_size_size == 0)
				return 0;
		}
	}

	// 1 is for the Misc field.
	return 1 + id_size + prop_size_size + prop_size;
}


/// \brief      "Encodes" the properties of the Sparse filter
static bool
properties_sparse(uint8_t *out, size_t *out_pos, size_t out_size,
		const lzma_options_sparse *options)
{
	if (options->size > out_size - *out_pos)
		return LZMA_BUF_ERROR;

	memcpy(out + *out_pos, options->bytes, options->size);
	*out_pos += options->size;

	return LZMA_OK;
}


/// \brief      Encodes LZMA Properties and Dictionary Flags (two bytes)
static bool
properties_lzma(uint8_t *out, size_t *out_pos, size_t out_size,
		const lzma_options_lzma *options)
{
	if (out_size - *out_pos < 2)
		return LZMA_BUF_ERROR;

	// LZMA Properties
	if (lzma_lzma_encode_properties(options, out + *out_pos))
		return LZMA_HEADER_ERROR;

	++*out_pos;

	// Dictionary flags
	//
	// Dictionary size is encoded using six bits of
	// which one is mantissa and five are exponent.
	//
	// There are some limits that must hold to keep
	// this coding working.
#	if LZMA_DICTIONARY_SIZE_MAX > UINT32_MAX / 2
#		error LZMA_DICTIONARY_SIZE_MAX is too big.
#	endif
#	if LZMA_DICTIONARY_SIZE_MIN < 1
#		error LZMA_DICTIONARY_SIZE_MIN cannot be zero.
#	endif

	// Validate it:
	if (options->dictionary_size < LZMA_DICTIONARY_SIZE_MIN
			|| options->dictionary_size > LZMA_DICTIONARY_SIZE_MAX)
		return LZMA_HEADER_ERROR;

	// Find the highest bit set:
	uint32_t highest = 1;
	while (options->dictionary_size >> highest)
		++highest;

	// Force rounding up: FIXME check that doc matches code
	//   ...0001xxxxx   <-- dict_size; 1 == highest bit set
	// + ...000011111   <-- forces rounding up
	// = ...00xxxxxxx   <-- rounded
	const uint32_t rounded = (UINT32_C(0x3FFFFFFF) >> (32 - highest))
			+ options->dictionary_size;

	// It is possible that rounding up caused the next
	// bit to get set. Fix value of highest if needed:
	if ((UINT32_C(1) << (highest + 1)) & rounded)
		++highest;

	// Encode the dictionary size.
	if (highest == 1) {
		// This is a special case.
		out[*out_pos] = 0x00;
	} else {
		// Set mantissa to 0 or 1 to indicate the
		// state of the second highest bit in rounded.
		const uint32_t mantissa
				= rounded & (UINT32_C(1) << (highest - 2));

		// Store the exponent and mantissa. We substract one
		// because when highest == 1 we don't have mantissa.
		// This way the maximum possible value is one step
		// bigger (2**32 instead of 2**31+2**30).
		out[*out_pos] = (2 * (highest - 1)) + mantissa - 1;
	}

	++*out_pos;

	return LZMA_OK;
}


extern LZMA_API lzma_ret
lzma_filter_flags_encode(uint8_t *out, size_t *out_pos, size_t out_size,
		const lzma_options_filter *options)
{
	// Minimum output is one byte (everything fits into Misc).
	if (*out_pos >= out_size)
		return LZMA_BUF_ERROR;

	// Get size of Filter Properties.
	const int32_t prop_size = get_properties_size(options);
	if (prop_size == -1)
		return LZMA_HEADER_ERROR;

	// Misc, External ID, and Size of Properties
	if (options->id < 0xE0
			&& (uint64_t)(prop_size) == options->id / 0x20) {
		// ID and Size of Filter Properties fit into Misc.
		out[*out_pos] = options->id;
		++*out_pos;

	} else if (prop_size <= 30) {
		// Size of Filter Properties fits into Misc.
		out[*out_pos] = prop_size + 0xE0;
		++*out_pos;

		// External ID is used to encode the Filter ID.
		size_t dummy = 0;
		switch (lzma_vli_encode(options->id, &dummy, 1,
				out, out_pos, out_size) != LZMA_STREAM_END) {
		case LZMA_STREAM_END:
			break;

		case LZMA_OK:
		case LZMA_BUF_ERROR:
			return LZMA_BUF_ERROR;

		default:
			return LZMA_HEADER_ERROR;
		}

	} else {
		// Nothing fits into Misc.
		out[*out_pos] = 0xFF;
		++*out_pos;

		// External ID is used to encode the Filter ID.
		size_t dummy = 0;
		switch (lzma_vli_encode(options->id, &dummy, 1,
				out, out_pos, out_size) != LZMA_STREAM_END) {
		case LZMA_STREAM_END:
			break;

		case LZMA_OK:
		case LZMA_BUF_ERROR:
			return LZMA_BUF_ERROR;

		default:
			return LZMA_HEADER_ERROR;
		}

		// External Size of Filter Properties
		dummy = 0;
		switch (lzma_vli_encode(prop_size, &dummy, 1,
				out, out_pos, out_size) != LZMA_STREAM_END) {
		case LZMA_STREAM_END:
			break;

		case LZMA_OK:
		case LZMA_BUF_ERROR:
			return LZMA_BUF_ERROR;

		default:
			return LZMA_HEADER_ERROR;
		}
	}

	// Filter Properties
	lzma_ret ret;
	switch (options->id) {
	case LZMA_FILTER_SPARSE:
		ret = properties_sparse(out, out_pos, out_size,
				options->options);
		break;

	case LZMA_FILTER_LZMA:
		ret = properties_lzma(out, out_pos, out_size,
				options->options);
		break;

	default:
		// Other filters don't have any properties.
		assert(prop_size == 0);
		ret = LZMA_OK;
		break;
	}

	return ret;
}
