///////////////////////////////////////////////////////////////////////////////
//
/// \file       raw_decoder.c
/// \brief      Raw decoder initialization API
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "raw_decoder.h"
#include "copy_coder.h"
#include "simple_coder.h"
#include "subblock_decoder.h"
#include "subblock_decoder_helper.h"
#include "lzma_decoder.h"
#include "metadata_decoder.h"


static lzma_init_function
get_function(uint64_t id)
{
	switch (id) {
#ifdef HAVE_FILTER_COPY
	case LZMA_FILTER_COPY:
		return &lzma_copy_decoder_init;
#endif

#ifdef HAVE_FILTER_SUBBLOCK
	case LZMA_FILTER_SUBBLOCK:
		return &lzma_subblock_decoder_init;
#endif

#ifdef HAVE_FILTER_SPARSE
#	error Sparse filter not implemented
	case LZMA_FILTER_SPARSE:
		assert(0);
#endif

#ifdef HAVE_FILTER_X86
	case LZMA_FILTER_X86:
		return &lzma_simple_x86_decoder_init;
#endif

#ifdef HAVE_FILTER_POWERPC
	case LZMA_FILTER_POWERPC:
		return &lzma_simple_powerpc_decoder_init;
#endif

#ifdef HAVE_FILTER_IA64
	case LZMA_FILTER_IA64:
		return &lzma_simple_ia64_decoder_init;
#endif

#ifdef HAVE_FILTER_ARM
	case LZMA_FILTER_ARM:
		return &lzma_simple_arm_decoder_init;
#endif

#ifdef HAVE_FILTER_ARMTHUMB
	case LZMA_FILTER_ARMTHUMB:
		return &lzma_simple_armthumb_decoder_init;
#endif

#ifdef HAVE_FILTER_SPARC
	case LZMA_FILTER_SPARC:
		return &lzma_simple_sparc_decoder_init;
#endif

#ifdef HAVE_FILTER_LZMA
	case LZMA_FILTER_LZMA:
		return &lzma_lzma_decoder_init;
#endif

#ifdef HAVE_FILTER_SUBBLOCK
	case LZMA_FILTER_SUBBLOCK_HELPER:
		return &lzma_subblock_decoder_helper_init;
#endif
	}

	return NULL;
}


extern lzma_ret
lzma_raw_decoder_init(lzma_next_coder *next, lzma_allocator *allocator,
		const lzma_options_filter *options,
		uint64_t uncompressed_size, bool implicit)
{
	return lzma_raw_coder_init(next, allocator, options, uncompressed_size,
			&get_function, implicit, false);
}


extern LZMA_API lzma_ret
lzma_raw_decoder(lzma_stream *strm, const lzma_options_filter *options,
		uint64_t uncompressed_size, lzma_bool implicit)
{
	if (options == NULL)
		return LZMA_STREAM_ERROR;

	lzma_ret ret;

	ret = lzma_strm_init(strm);
	if (ret != LZMA_OK)
		return ret;

	strm->internal->supported_actions[LZMA_RUN] = true;
	strm->internal->supported_actions[LZMA_FLUSH] = true;

	ret = lzma_raw_decoder_init(&strm->internal->next, strm->allocator,
			options, uncompressed_size, implicit);

	if (ret != LZMA_OK)
		(void)lzma_end(strm);

	return ret;
}
