///////////////////////////////////////////////////////////////////////////////
//
/// \file       stream_encoder_single.c
/// \brief      Encodes Single-Block .lzma files
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "stream_common.h"
#include "block_encoder.h"


struct lzma_coder_s {
	lzma_next_coder next;

	// Uncompressed Size, Backward Size, and Footer Magic Bytes are
	// part of Block in the file format specification, but it is simpler
	// to implement them as part of Stream.
	enum {
		SEQ_STREAM_HEADER_COPY,
		SEQ_DATA_CODE,
		SEQ_STREAM_FOOTER_INIT,
		SEQ_STREAM_FOOTER_COPY,
	} sequence;

	size_t pos;

	// Stream Header or Footer
	uint8_t *header;
	size_t header_size;

	lzma_stream_flags stream_flags;
};


static lzma_ret
stream_encode(lzma_coder *coder, lzma_allocator *allocator,
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out, size_t *out_pos,
		size_t out_size, lzma_action action)
{
	while (*out_pos < out_size)
	switch (coder->sequence) {
	case SEQ_STREAM_HEADER_COPY:
		bufcpy(coder->header, &coder->pos, coder->header_size,
				out, out_pos, out_size);

		if (coder->pos == coder->header_size) {
			lzma_free(coder->header, allocator);
			coder->header = NULL;

			coder->pos = 0;
			coder->sequence = SEQ_DATA_CODE;
		}

		break;

	case SEQ_DATA_CODE: {
		// Store the current positions so that we can update
		// coder->uncompressed_size and coder->backward_size.
		const size_t in_start = *in_pos;
		const size_t out_start = *out_pos;

		// Encode
		const lzma_ret ret = coder->next.code(coder->next.coder,
				allocator, in, in_pos, in_size,
				out, out_pos, out_size, action);

		// Update the sizes.
		coder->stream_flags.uncompressed_size += *in_pos - in_start;
		coder->stream_flags.backward_size += *out_pos - out_start;

		// Check the return value of the encoder.
		if (ret != LZMA_STREAM_END)
			return ret;

		assert(*in_pos == in_size);

		coder->sequence = SEQ_STREAM_FOOTER_INIT;
		break;
	}

	case SEQ_STREAM_FOOTER_INIT: {
		const lzma_ret ret = lzma_stream_footer_encode(
				&coder->header, &coder->header_size,
				allocator, &coder->stream_flags);
		if (ret != LZMA_OK)
			return ret;
	}

	// Fall through

	case SEQ_STREAM_FOOTER_COPY:
		bufcpy(coder->header, &coder->pos, coder->header_size,
				out, out_pos, out_size);
		if (coder->pos == coder->header_size)
			return LZMA_STREAM_END;
		break;

	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}


static void
stream_encoder_end(lzma_coder *coder, lzma_allocator *allocator)
{
	lzma_next_coder_end(&coder->next, allocator);
	lzma_free(coder, allocator);
	return;
}


extern lzma_ret
lzma_stream_encoder_single_init(lzma_next_coder *next,
		lzma_allocator *allocator, const lzma_options_stream *options)
{
	if (next->init != &lzma_stream_encoder_single_init) {
		// Free the old coder.
		lzma_next_coder_end(next, allocator);

		// Allocate a new Single-Block Stream encoder structure.
		next->coder = lzma_alloc(sizeof(lzma_coder), allocator);
		if (next->coder == NULL)
			return LZMA_MEM_ERROR;

		// Mark that we haven't initialized any filters yet.
		next->coder->next = LZMA_NEXT_CODER_INIT;
	}

	// Initialize stream_flags and encode Stream Header
	{
		next->coder->stream_flags = (lzma_stream_flags){
			.check = options->check,
			.has_crc32 = options->has_crc32,
			.is_multi = false,
			.uncompressed_size = 0,
			.backward_size = 0,

		};
		const lzma_ret ret = lzma_stream_header_encode(
				&next->coder->header,
				&next->coder->header_size,
				allocator, &next->coder->stream_flags);
		if (ret != LZMA_OK) {
			stream_encoder_end(next->coder, allocator);
			return ret;
		}
	}

	// Initialize Block encoder.
	{
		lzma_options_block block = {
			.check = options->check,
			.has_crc32 = options->has_crc32,
			.uncompressed_size = UINT64_MAX,
			.compressed_size = UINT64_MAX,
			.has_eopm = false,
			.is_metadata = false,
			.alignment = options->alignment
					+ next->coder->header_size,
		};
		memcpy(block.filters, options->filters,
				sizeof(options->filters));

		const lzma_ret ret = lzma_block_encoder_init(
				&next->coder->next, allocator, &block);
		if (ret != LZMA_OK) {
			stream_encoder_end(next->coder, allocator);
			return ret;
		}
	}

	// Reset the variables.
	next->coder->sequence = SEQ_STREAM_HEADER_COPY;
	next->coder->pos = 0;

	// All successfully initialized, set the function pointers.
	next->init = &lzma_stream_encoder_single_init;
	next->code = &stream_encode;
	next->end = &stream_encoder_end;

	return LZMA_OK;
}


extern LZMA_API lzma_ret
lzma_stream_encoder_single(
		lzma_stream *strm, const lzma_options_stream *options)
{
	if (options == NULL)
		return LZMA_STREAM_ERROR;

	lzma_ret ret;

	ret = lzma_strm_init(strm);
	if (ret != LZMA_OK)
		return ret;

	strm->internal->supported_actions[LZMA_RUN] = true;
	strm->internal->supported_actions[LZMA_FINISH] = true;

	ret = lzma_stream_encoder_single_init(&strm->internal->next,
			strm->allocator, options);

	if (ret != LZMA_OK)
		(void)lzma_end(strm);

	return ret;
}
