///////////////////////////////////////////////////////////////////////////////
//
/// \file       stream_flags_decoder.c
/// \brief      Decodes Stream Header and Footer from .lzma files
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "stream_flags_decoder.h"
#include "stream_common.h"


////////////
// Common //
////////////

struct lzma_coder_s {
	enum {
		SEQ_HEADER_MAGIC,
		SEQ_HEADER_FLAGS,
		SEQ_HEADER_CRC32,

		SEQ_FOOTER_UNCOMPRESSED_SIZE,
		SEQ_FOOTER_BACKWARD_SIZE,
		SEQ_FOOTER_FLAGS,
		SEQ_FOOTER_MAGIC,
	} sequence;

	size_t pos;
	uint32_t crc32;

	lzma_stream_flags *options;

};


static void
stream_header_decoder_end(lzma_coder *coder, lzma_allocator *allocator)
{
	lzma_free(coder, allocator);
	return;
}


static bool
stream_flags_decode(const uint8_t *in, lzma_stream_flags *options)
{
	// Reserved bits must be unset.
	if (*in & 0xE0)
		return true;

	options->check = *in & 0x07;
	options->has_crc32 = (*in & 0x08) != 0;
	options->is_multi = (*in & 0x10) != 0;

	return false;
}


////////////
// Header //
////////////

static lzma_ret
stream_header_decode(lzma_coder *coder,
		lzma_allocator *allocator lzma_attribute((unused)),
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out lzma_attribute((unused)),
		size_t *restrict out_pos lzma_attribute((unused)),
		size_t out_size lzma_attribute((unused)),
		lzma_action action lzma_attribute((unused)))
{
	while (*in_pos < in_size)
	switch (coder->sequence) {
	case SEQ_HEADER_MAGIC:
		if (in[*in_pos] != lzma_header_magic[coder->pos])
			return LZMA_DATA_ERROR;

		++*in_pos;

		if (++coder->pos == sizeof(lzma_header_magic)) {
			coder->pos = 0;
			coder->sequence = SEQ_HEADER_FLAGS;
		}

		break;

	case SEQ_HEADER_FLAGS:
		if (stream_flags_decode(in + *in_pos, coder->options))
			return LZMA_HEADER_ERROR;

		coder->crc32 = lzma_crc32(in + *in_pos, 1, 0);

		++*in_pos;
		coder->sequence = SEQ_HEADER_CRC32;
		break;

	case SEQ_HEADER_CRC32:
		if (in[*in_pos] != ((coder->crc32 >> (coder->pos * 8)) & 0xFF))
			return LZMA_DATA_ERROR;

		++*in_pos;

		if (++coder->pos == 4)
			return LZMA_STREAM_END;

		break;

	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}


extern lzma_ret
lzma_stream_header_decoder_init(lzma_next_coder *next,
		lzma_allocator *allocator, lzma_stream_flags *options)
{
	if (next->init != &lzma_stream_header_decoder_init) {
		lzma_next_coder_end(next, allocator);

		next->coder = lzma_alloc(sizeof(lzma_coder), allocator);
		if (next->coder == NULL)
			return LZMA_MEM_ERROR;
	}

	next->coder->sequence = SEQ_HEADER_MAGIC;
	next->coder->pos = 0;
	next->coder->crc32 = 0;
	next->coder->options = options;

	next->init = &lzma_stream_header_decoder_init;
	next->code = &stream_header_decode;
	next->end = &stream_header_decoder_end;

	return LZMA_OK;
}


extern LZMA_API lzma_ret
lzma_stream_header_decoder(lzma_stream *strm, lzma_stream_flags *options)
{
	if (options == NULL)
		return LZMA_STREAM_ERROR;

	lzma_ret ret;

	ret = lzma_strm_init(strm);
	if (ret != LZMA_OK)
		return ret;

	strm->internal->supported_actions[LZMA_FINISH] = true;

	ret = lzma_stream_header_decoder_init(&strm->internal->next,
			strm->allocator, options);

	if (ret != LZMA_OK)
		(void)lzma_end(strm);

	return ret;
}


////////////
// Footer //
////////////

static lzma_ret
stream_footer_decode(lzma_coder *coder,
		lzma_allocator *allocator lzma_attribute((unused)),
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out lzma_attribute((unused)),
		size_t *restrict out_pos lzma_attribute((unused)),
		size_t out_size lzma_attribute((unused)),
		lzma_action action lzma_attribute((unused)))
{
	while (*in_pos < in_size)
	switch (coder->sequence) {
	case SEQ_FOOTER_UNCOMPRESSED_SIZE: {
		const lzma_ret ret = lzma_vli_decode(
				&coder->options->uncompressed_size,
				&coder->pos, in, in_pos, in_size);
		if (ret != LZMA_STREAM_END)
			return ret;

		coder->pos = 0;
		coder->sequence = SEQ_FOOTER_BACKWARD_SIZE;
		break;
	}

	case SEQ_FOOTER_BACKWARD_SIZE: {
		const lzma_ret ret = lzma_vli_decode(
				&coder->options->backward_size, &coder->pos,
				in, in_pos, in_size);
		if (ret != LZMA_STREAM_END)
			return ret;

		coder->pos = 0;
		coder->sequence = SEQ_FOOTER_FLAGS;
		break;
	}

	case SEQ_FOOTER_FLAGS:
		if (stream_flags_decode(in + *in_pos, coder->options))
			return LZMA_HEADER_ERROR;

		++*in_pos;
		coder->sequence = SEQ_FOOTER_MAGIC;
		break;

	case SEQ_FOOTER_MAGIC:
		if (in[*in_pos] != lzma_footer_magic[coder->pos])
			return LZMA_DATA_ERROR;

		++*in_pos;

		if (++coder->pos == sizeof(lzma_footer_magic))
			return LZMA_STREAM_END;

		break;

	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}


extern lzma_ret
lzma_stream_footer_decoder_init(lzma_next_coder *next,
		lzma_allocator *allocator, lzma_stream_flags *options,
		lzma_bool decode_uncompressed_size)
{
	// NOTE: We compare to *header* decoder init function, because
	// both use the same data structure. Avoids useless free+alloc
	// when the same lzma_next_coder is used first for header and
	// then for footer or vice versa.
	if (next->init != &lzma_stream_header_decoder_init) {
		lzma_next_coder_end(next, allocator);

		next->coder = lzma_alloc(sizeof(lzma_coder), allocator);
		if (next->coder == NULL)
			return LZMA_MEM_ERROR;
	}

	if (decode_uncompressed_size) {
		options->uncompressed_size = 0;
		next->coder->sequence = SEQ_FOOTER_UNCOMPRESSED_SIZE;
	} else {
		next->coder->sequence = SEQ_FOOTER_BACKWARD_SIZE;
	}

	options->backward_size = 0;
	next->coder->pos = 0;
	next->coder->options = options;

	// See the comments in the beginning of this function. For the same
	// reason, we can use the same end function as the header decoder.
	next->init = &lzma_stream_header_decoder_init;
	next->code = &stream_footer_decode;
	next->end = &stream_header_decoder_end;

	return LZMA_OK;
}


extern LZMA_API lzma_ret
lzma_stream_footer_decoder(lzma_stream *strm, lzma_stream_flags *options,
		lzma_bool decode_uncompressed_size)
{
	if (options == NULL)
		return LZMA_STREAM_ERROR;

	lzma_ret ret;

	ret = lzma_strm_init(strm);
	if (ret != LZMA_OK)
		return ret;

	strm->internal->supported_actions[LZMA_RUN] = true;

	ret = lzma_stream_footer_decoder_init(&strm->internal->next,
			strm->allocator, options, decode_uncompressed_size);

	if (ret != LZMA_OK)
		(void)lzma_end(strm);

	return ret;
}


extern LZMA_API lzma_ret
lzma_stream_footer_reverse_decode(const uint8_t *in, size_t *in_size,
		lzma_stream_flags *options)
{
	lzma_coder coder = {
		.sequence = SEQ_FOOTER_MAGIC,
		.pos = 1,
	};

	while (*in_size != 0)
	switch (coder.sequence) {
	case SEQ_FOOTER_UNCOMPRESSED_SIZE: {
		// We don't want to update *in_size even if decoding succeeds.
		size_t tmp = *in_size;
		const lzma_ret ret = lzma_vli_reverse_decode(
				&options->uncompressed_size, in, &tmp);

		// If decoding fails, it's not fatal. Just mark that
		// uncompressed_size is unknown.
		if (ret != LZMA_OK)
			options->uncompressed_size = UINT64_MAX;

		return LZMA_OK;
	}

	case SEQ_FOOTER_BACKWARD_SIZE: {
		const lzma_ret ret = lzma_vli_reverse_decode(
				&options->backward_size, in, in_size);
		if (ret != LZMA_OK)
			return ret;

		coder.sequence = SEQ_FOOTER_UNCOMPRESSED_SIZE;
		break;

	}

	case SEQ_FOOTER_FLAGS:
		if (stream_flags_decode(in + *in_size - 1, options))
			return LZMA_HEADER_ERROR;

		--*in_size;
		coder.sequence = SEQ_FOOTER_BACKWARD_SIZE;
		break;

	case SEQ_FOOTER_MAGIC:
		if (in[*in_size] != lzma_footer_magic[coder.pos])
			return LZMA_DATA_ERROR;

		if (coder.pos == 0)
			coder.sequence = SEQ_FOOTER_FLAGS;

		--coder.pos;
		--*in_size;
		break;

	default:
		assert(false);
		return LZMA_STREAM_ERROR;
	}

	// It's OK if we couldn't parse uncompressed_size.
	if (coder.sequence == SEQ_FOOTER_UNCOMPRESSED_SIZE) {
		options->uncompressed_size = UINT64_MAX;
		return LZMA_OK;
	}

	return LZMA_BUF_ERROR;
}
